__all__ = ['RigolTechnologiesMSO4014', 'TektronixMSO64B', 'TektronixMSO64BSpectrogram']

import labbench as lb
from labbench import paramattr as attr

scope_channel_kwarg = attr.method_kwarg.int(
    'channel', min=1, max=4, help='hardware input port'
)


@scope_channel_kwarg
class RigolTechnologiesMSO4014(lb.VISADevice):
    # used for automatic connection
    make = attr.value.str('RIGOL TECHNOLOGIES', inherit=True)
    model = attr.value.str('MSO4014', inherit=True)

    time_offset = attr.property.float(
        key=':TIM:OFFS',
        label='s',
        help='acquisition time offset relative to the trigger',
    )
    time_scale = attr.property.float(
        key=':TIM:SCAL', label='s', help='acquisition time per division'
    )
    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    def open(self, horizontal=False):
        self.write(':WAVeform:FORMat ASCii')

    def fetch(self):
        return self.backend.query_ascii_values(':WAV:DATA?')

    def fetch_rms(self):
        return float(self.backend.query(':MEAS:VRMS?').rstrip().lstrip())


@attr.visa_keying(remap={True: '1', False: '0'})
@scope_channel_kwarg
class TektronixMSO64B(lb.VISADevice):
    # used for automatic connection
    make = attr.value.str('TEKTRONIX', inherit=True)
    model = attr.value.str('MSO64B', inherit=True)
    timeout: float = attr.value.float(3)

    # horizontal acquisition
    record_length = attr.property.int(
        key='HORIZONTAL:RECORDLENGTH', label='samples', help='acquisition record length'
    )
    sample_rate = attr.property.float(key='HOR:SAMPLER', label='S/s', get_on_set=True)
    _horizontal_mode = attr.property.str(
        key='HORizontal:MODe',
        case=False,
        only=('AUTOMATIC', 'MANUAL'),
        help='AUTO automatically adjusts the sample rate and record length',
    )

    # input port parameters
    # TODO: check lower bound
    bandwidth = attr.method.float(
        key='CH{channel}:BANDWIDTH',
        min=0,
        max=8e9,
        label='Hz',
        help='input port analog bandwidth',
    )
    channel_enabled = attr.method.bool(
        key='SELect:CH{channel}',
        help='enable channel acquisition',
    )
    coupling = attr.method.str(
        key='CH{channel}:COUPLING',
        case=False,
        only=('AC', 'DC', 'DCReject'),
        cache=True,
        help='input port coupling mode',
    )
    input_termination = attr.method.float(
        key='CH{channel}:TERMINATION',
        only=(50, 1e6),
        cache=True,
        help='input port termination impedance',
    )

    # vertical acquisition
    vertical_scale = attr.property.float(
        key='CH{channel}:SCALE',
        label='V',
        help='vertical scale of the specified channel',
    )

    def open(self):
        self._horizontal_mode = 'MANUAL'
        self.write(':VERBose 0')
        self.write(':HEADer 0')


class TektronixMSO64BSpectrogram(TektronixMSO64B):
    resolution_bandwidth = attr.property.float(
        key='SV:RBW', label='Hz', help='resolution bandwidth of all channels'
    )
    span = attr.property.float(
        key='SV:SPAN', max=2e9, label='Hz', help='analysis bandwidth per channel'
    )

    # don't log instrument front-end display settings that don't impact the data (log=False)
    _power_scale_min = attr.property.float(
        key='SV:SPECtrogram:CSCale:MIN',
        min=-170,
        max=99,
        log=False,
        label='dBm',
        help='spectrogram display power scale minimum',
    )
    _power_scale_max = attr.property.float(
        key='SV:SPECtrogram:CSCale:MAX',
        min=-170,
        max=99,
        log=False,
        label='dBm',
        help='spectrogram display power scale maximum',
    )

    # assume that these are only set once, so relegate them to metadata (cache=True)
    _sync_center_frequencies = attr.property.bool(
        key='SV:LOCKCenter',
        cache=True,
        help='whether to lock all channels to the same center frequency',
    )
    _resolution_bandwidth_mode = attr.property.str(
        key='SV:RBWMode',
        case=False,
        only=('AUTOMATIC', 'MANUAL'),
        cache=True,
        help='whether to automatically set analysis parameters to achieve a specified RBW',
    )

    center_frequency = attr.method.float(
        key='CH{channel}:SV:CENTERFrequency',
        max=8e9,
        label='Hz',
        help='channel center frequency',
    )

    bandwidth = attr.method.float(
        # TODO: check min
        key='CH{channel}:BANDWIDTH',
        min=0,
        max=8e9,
        label='Hz',
        help='channel analysis bandwidth',
    )

    # log=False here since we intend usage of spectrogram_enabled
    _only_spectrogram_enabled = attr.method.bool(
        key='SV:CH{channel}:SEL:SPEC',
        log=False,
        help='enable channel spectrogram analysis',
    )

    _only_spectrumview_enabled = attr.method.bool(
        key='CH{channel}:SV:STATE',
        log=False,
        help='enable channel spectrum view mode',
    )

    @attr.method.bool(help='channel center frequency')
    @scope_channel_kwarg
    def spectrogram_enabled(self, enabled: bool = lb.Undefined, /, *, channel: int):
        return all(
            [
                self.channel_enabled(enabled, channel),
                self._only_spectrumview_enabled(enabled, channel),
                self._only_spectrogram_enabled(enabled, channel),
            ]
        )

    def open(self):
        self._resolution_bandwidth_mode = 'MANUAL'
        self._sync_center_frequencies = False


if __name__ == '__main__':
    with TektronixMSO64BSpectrogram() as scope:
        scope.spectrogram_enabled(channel=1)

        scope.resolution_bandwidth = 10e3
        scope.span = 2e9

        scope.coupling('AC', channel=1)
        scope.bandwidth(8e9, channel=1)
        scope.input_termination(50, channel=1)
        scope.center_frequency(3e9, channel=1)
        scope.vertical_scale(0.05, channel=1)



#!/usr/bin/env python3

"""
Created on Wed Feb  7 10:05:35 2018

@author: aec

This is where we drive things that move: 
    for now this is just the ETS-Lindgren model 2005 Azimuth Positioner
"""

import labbench as lb
from labbench import paramattr as attr

__all__ = ['ETSLindgrenAzi2005']


@attr.visa_keying(write_fmt='{key}{value}')
class ETSLindgrenAzi2005(lb.VISADevice):
    # constructor argument fields
    timeout: float = attr.value.float(default=20, min=0, label='s')
    baud_rate: int = attr.value.int(default=9600, min=1, label='baud')
    parity = attr.value.bytes(default=b'N')
    stopbits = attr.value.float(default=1, min=1, max=2, step=0.5)
    xonxoff = attr.value.bool(default=False)
    rtscts = attr.value.bool(default=False)
    dsrdtr = attr.value.bool(default=False)

    read_termination: str = attr.value.str('\n', inherit=True)
    write_termination: str = attr.value.str('\r', inherit=True)

    def config(self, mode):
        if mode in ('CR' or 'NCR'):
            self.write(self, mode)
        else:
            print('check your spelling!')

    def whereami(self):
        return self.query('CP?')

    def wheredoigo(self):
        return self.query('DIR?')

    def set_speed(self, value):
        self.write('S' + value)

    # return self.query_speed

    def set_limits(self, side, value):
        """Probably should put some error checking in here to make sure value is a float
        Also, note we use write here becuase property.setter inserts a space"""
        if side == 'lower':
            self.write('LL' + value)
        elif side == 'upper':
            self.write('UL' + value)
        else:
            print('you typed the wrong thing.')

    def set_position(self, value):
        self.write('CP' + value)
        """Important note: 
            This command labels the position with a value. It does not move the turntable!"""

    def seek(self, value):
        self.write('SK' + value)

    def stop(self):
        self.write('ST')
        lb.sleep(3)
        print(self.wheredoigo())
        """If wheredoigo returns N, we are stopped"""
        # didistop = self.wheredoigo()
        # if didistop[0] == 'N':
        #    print('yay we stopped!')
        # else:
        #   print('oops still moving!')

    # A bunch of command-keyed states
    speed = attr.property.int(key='S', min=0, max=3, help='speed')
    cwlimit = attr.property.float(
        key='UL', min=000.0, max=999.9, step=0.1, help='cwlimit'
    )
    cclimit = attr.property.float(
        key='LL', min=000.0, max=999.9, step=0.1, help='cclimit'
    )
    define_position = attr.property.float(
        key='CP', min=0, max=360, step=0.1, help='rotation (degrees)'
    )
    position = attr.property.float(
        key='SK', min=0, max=360, help='rotation (degrees)', gets=False
    )


if __name__ == '__main__':
    # Enable labbench debug messages
    # log_to_screen()
    lb.show_messages('debug')

    with ETSLindgrenAzi2005('COM4') as motor:
        #        print(motor.query('?'))
        motor.set_position('30')
        print(repr(motor.whereami()))
        # Configure
#        print(motor.define_position)
#        motor.position = 30
#        motor.write('*WAI')
#        print(motor.define_position)
# print(motor.position)



"""
Drivers for USB peripherals

:author: Dan Kuester <daniel.kuester@nist.gov>, Andre Rosete <andre.rosete@nist.gov>
"""

import collections
import labbench as lb
from labbench import paramattr as attr


class AcronamePropertyAdapter(attr.KeyAdapterBase):
    def set(self, device, key: tuple, value, trait=None):
        """Apply an instrument setting to the instrument. The value ``value''
        will be applied to the trait attriute ``attr'' in type(self).
        """
        op, channel = key

        if op == 'data':
            device._set_data_enabled(channel, value)
        elif op == 'power':
            device._set_power_enabled(channel, value)
        else:
            raise ValueError(
                'first element in command key must be one of "data" or "power"'
            )

    def get(self, device, key: tuple, trait=None):
        """Apply an instrument setting to the instrument. The value ``value''
        will be applied to the trait attriute ``attr'' in type(self).
        """
        op, channel = key

        if op == 'data':
            device._get_data_enabled(channel)
        elif op == 'power':
            device._get_power_enabled(channel)
        else:
            raise ValueError(
                'first element in command key must be one of "data" or "power"'
            )


@AcronamePropertyAdapter()
class AcronameUSBHub2x4(lb.Device):
    """A USB hub with control over each port."""

    MODEL = 17

    def _set_power_enable(self, port: int, enabled: bool):
        if enabled:
            self._hub.usb.setPortEnable(port)
        else:
            self._hub.usb.setPortDisable(port)

    def _set_data_enable(self, port: int, enabled: bool):
        if enabled:
            self._hub.usb.setDataEnable(port)
        else:
            self._hub.usb.setDataDisable(port)

    def _get_power_enable(self, port: int):
        self._hub.usb.getPortEnable(port)

    def _get_data_enable(self, port: int):
        return self._hub.usb.getDataEnable(port)

    data0_enabled = attr.property.bool(key=('data', 0))
    data1_enabled = attr.property.bool(key=('data', 1))
    data2_enabled = attr.property.bool(key=('data', 2))
    data3_enabled = attr.property.bool(key=('data', 3))

    power0_enabled = attr.property.bool(key=('power', 0))
    power1_enabled = attr.property.bool(key=('power', 1))
    power2_enabled = attr.property.bool(key=('power', 2))
    power3_enabled = attr.property.bool(key=('power', 3))

    resource: str = attr.value.str(
        allow_none=True,
        cache=True,
        help='None to autodetect, or a serial number string',
    )

    def open(self):
        import brainstem

        specs = self._bs.discover.findAllModules(brainstem.link.Spec.USB)

        specs = [s for s in specs if s.model == self.MODEL]
        if self.resource is not None:
            specs = [s for s in specs if s.serial_number == self.resource]

        if len(specs) > 1:
            if self.resource is None:
                raise Exception(
                    'More than one connected USB device matches model '
                    + str(self.MODEL)
                    + ' - provide serial number?'
                )
            else:
                raise Exception(
                    'More than one connected USB device match model '
                    + str(self.MODEL)
                    + ' and serial '
                    + str(self.resource)
                )
        elif len(specs) == 0:
            raise Exception(
                'No USB devices connected that match model '
                + str(self.MODEL)
                + ' and serial '
                + str(self.resource)
            )

        self.backend = brainstem.stem.USBHub2x4()
        self.backend.connectFromSpec(specs[0])

    def close(self):
        """Release control over the device."""
        self.backend.disconnect()

    def _set_power_enabled(self, channel, enabled):
        """Enable or disable of USB port features at one or all hub ports.

        :param data:  Enables data on the port (if evaluates to true)

        :param power: Enables power on the port (if evaluates to true)

        :param channel: An integer port number specifies the port to act on,
         otherwise 'all' (the default) applies the port settings to all
         ports on the hub.
        """
        if channel == 'all':
            channels = list(range(4))
        elif isinstance(channel, collections.Iterable):
            channels = channel
        else:
            channels = [channel]

        for chan in channels:
            if enabled:
                self._hub.usb.setPortEnable(chan)
            else:
                self._hub.usb.setPortDisable(chan)

    def _set_data_enabled(self, channel, enabled):
        """Enable or disable of USB port features at one or all hub ports.

        :param data:  Enables data on the port (if evaluates to true)

        :param power: Enables power on the port (if evaluates to true)

        :param channel: An integer port number specifies the port to act on,
         otherwise 'all' (the default) applies the port settings to all
         ports on the hub.
        """
        if channel == 'all':
            channels = list(range(4))
        elif isinstance(channel, collections.Iterable):
            channels = channel
        else:
            channels = [channel]

        for chan in channels:
            if enabled:
                self._hub.usb.setDataEnable(chan)
            else:
                self._hub.usb.setDataDisable(chan)



import time
import labbench as lb
from labbench import paramattr as attr
import platform
import numpy as np
from threading import Lock

__all__ = ['MiniCircuitsUSBDevice', 'SwitchAttenuatorBase']

usb_enumerate_lock = Lock()
usb_command_lock = Lock()
usb_registry = {}  # serial number: USB path


class MiniCircuitsUSBDevice(lb.Device):
    """General control over MiniCircuits USB devices"""

    _VID = 0x20CE  # USB HID Vendor ID

    resource: str = attr.value.str(
        default=None,
        help='serial number; must be set if more than one device is connected',
        cache=True,
    )

    # annotated values can be passed as constructor arguments
    usb_path: bytes = attr.value.bytes(
        default=None,
        help='if not None, override `resource` to connect to a specific USB path',
        cache=True,
    )

    timeout: float = attr.value.float(default=1, min=0.5, label='s', cache=True)

    def open(self):
        import hid

        if self.usb_path is None:
            self.usb_path = self._find_path(self.resource)

        self.backend = hid.device()
        self.backend.open_path(self.usb_path)
        self.backend.set_nonblocking(1)

        usb_registry[self.usb_path] = self.serial_number

        if self.usb_path is None:
            self._logger.info(
                'connected to {self.model} with serial {self.serial_number}'
            )

    def close(self):
        if self.backend:
            self.backend.close()

    @classmethod
    def _parse_str(cls, data):
        """Convert a command response to a string."""
        b = np.array(data[1:], dtype='uint8').tobytes()
        return b.split(b'\x00', 1)[0].decode()

    def _cmd(self, *cmd):
        """Send up to 64 1-byte unsigned integers and return the response."""
        with usb_command_lock:
            if len(cmd) > 64:
                raise ValueError('command key data length is limited to 64')

            cmd = list(cmd) + (63 - len(cmd)) * [0]

            if platform.system().lower() == 'windows':
                self.backend.write([0] + cmd[:-1])
            else:
                self.backend.write(cmd)

            t0 = time.time()
            msg = None
            while time.time() - t0 < self.timeout:
                d = self.backend.read(64)
                if d:
                    if d[0] == cmd[0]:
                        break
                    else:
                        msg = 'device responded to command code {}, but expected {} (full response {})'.format(
                            d[0], cmd[0], repr(d)
                        )
            else:
                if msg is None:
                    raise TimeoutError('no response from device')
                else:
                    raise lb.DeviceException(msg)

        return d

    @classmethod
    def _test_instance(cls, usb_path):
        """must return a trial object to test connections when enumerating devices.
        the subclass must have serial_number and model traits.
        """
        raise NotImplementedError(
            'subclasses must implement this to return an instance for trial connection'
        )

    @classmethod
    def _find_path(cls, serial):
        """Find a USB HID device path matching the MiniCircuits device with
        the specified serial number. If serial is None, then check that
        exactly one MiniCircuits device is connected, and return its path.
        Raise an exception if no devices are connected.
        """
        import hid
        unknown_match = []

        with usb_enumerate_lock:
            found = {}

            for dev in hid.enumerate(cls._VID, cls._PID):
                # Check for a cached serial number first
                try:
                    this_serial = usb_registry[dev['path']]
                    found[this_serial] = dev['path']
                    continue
                except KeyError:
                    pass

                try:
                    # serial number is unfamiliar - probe the device
                    with cls._test_instance(dev['path']) as inst:
                        this_serial = inst.serial_number
                        usb_registry[dev['path']] = this_serial
                        found[this_serial] = dev['path']

                except OSError:
                    # potentially open in another process
                    unknown_match.append(dev)
                    continue

        if len(found) == 0:
            ex = ConnectionError(
                f'found no USB HID devices matching the vendor and product of {cls.__qualname__}'
            )

            if len(unknown_match) > 0 and hasattr(ex, 'add_note'):
                # python>=3.10
                ex.add_note(f'hid failed to open these matches to vid={hex(cls._VID)} pid={hex(cls._PID)}:')
                for dev in unknown_match:
                    ex.add_note(f'\t{dev["path"]}')

            raise ex

        names = ', '.join([repr(k) for k in found.keys()])

        if serial is None:
            if len(found) == 1:
                ret = list(found.values())[0]
                return ret
            else:
                raise ConnectionError(
                    f'specify one of the available {cls.__name__} resources: {names}'
                )

        try:
            ret = found[serial]
        except KeyError:
            raise ConnectionError(
                f'specified resource {repr(serial)}, but only {names} are available'
            )
        return ret


# class PowerSensor(MiniCircuitsUSBDevice):
#    """
#    Class for interfacing with Mini-Circuits USB Power Sensors.
#    Verified working:
#        - PWR-6GHS
#    """
#
#    # Mini-Circuits PWR-6GHS USB Power Sensor
#    PID = 0x11
#
#    # Not sure if any of these command codes are also shared.
#    CMD_GET_MODEL_NAME = 104
#    CMD_GET_SERIAL = 105
#    CMD_SET_MEASUREMENT_MODE = 15
#    CMD_READ_POWER = 102
#    CMD_GET_TEMPERATURE = 103
#    CMD_GET_FIRMWARE_VERSION = 99
#
#    def get_model_name(self):
#        d = self._cmd(self.CMD_GET_MODEL_NAME)
#        return self._parse_str(d)
#
#    def get_serial(self):
#        d = self._cmd(self.CMD_GET_SERIAL)
#        return self._parse_str(d)
#
#    def set_measurement_mode(self, mode='low-noise'):
#        mode_num = {
#            'low-noise': 0,
#            'fast-sampling': 1,
#            'fastest-sampling': 2,
#        }[mode]
#        self._cmd(self.CMD_SET_MEASUREMENT_MODE, mode_num)
#
#    def get_power(self, freq):
#        if freq > 65.535e6:
#            scale = 1e6
#            units = 'M'
#        else:
#            scale = 1e3
#            units = 'k'
#        freq = int(round(freq / scale))
#        freq1 = freq >> 8
#        freq2 = freq - (freq1 << 8)
#        d = self._cmd(self.CMD_READ_POWER, freq1, freq2, ord(units))
#        s = ''.join(chr(c) for c in d[1:7])
#        return float(s)
#
#    def get_temperature(self):
#        d = self._cmd(self.CMD_GET_TEMPERATURE)
#        s = ''.join(chr(c) for c in d[1:7])
#        return float(s)
#
#    def get_firmware_version(self):
#        d = self._cmd(self.CMD_GET_FIRMWARE_VERSION)
#        return chr(d[5]) + chr(d[6])


class SwitchAttenuatorBase(MiniCircuitsUSBDevice):
    CMD_GET_PART_NUMBER = 40
    CMD_GET_SERIAL_NUMBER = 41

    @classmethod
    def _test_instance(cls, usb_path):
        device = SwitchAttenuatorBase(usb_path=usb_path)
        device._logger.logger.disabled = True
        return device

    @attr.property.str(sets=False, cache=True)
    def model(self):
        d = self._cmd(self.CMD_GET_PART_NUMBER)
        return self._parse_str(d)

    @attr.property.str(sets=False, cache=True)
    def serial_number(self):
        d = self._cmd(self.CMD_GET_SERIAL_NUMBER)
        return self._parse_str(d)



""" GPS simulator control classes

:author: Duncan McGillivray <duncan.mcgillivray@nist.gov>, Daniel Kuester <daniel.kuester@nist.gov>
"""

__all__ = ['SpirentGSS8000']

import time
import re
import labbench as lb
from labbench import paramattr as attr

status_messages = (
    b'no scenario',
    b'loading',
    b'ready',
    b'arming',
    b'armed',
    b'running',
    b'paused',
    b'ended',
)


class SpirentGSS8000(lb.SerialDevice):
    """Control a Spirent GPS GSS8000 simulator over a serial connection.

    Responses from the Spirent seem to be incompatible with
    pyvisa, so this driver uses plain serial.
    """

    resource = attr.value.str(
        help="serial port address ('COM<nn>' in windows, '/dev/<xxxx>' in linux, etc.)",
    )

    def get_key(self, key, trait_name=None):
        "Status messages that may be received from the instrument"
        return self.query(key)

    def load_scenario(self, path):
        """Load a GPS scenario from a file stored on the instrument.

        :param path: Full path to scenario file on the instrument.

        """
        # write scenario path as 'c:\folder\folder\name.scn'
        loadpath = self.fix_path_name(path)
        self.backend.read(self.backend.inWaiting())
        self.backend.write('SC,%s\n' % (loadpath))
        self.current_scenario

    def save_scenario(self, folderpath: str):
        """Save the current GPS scenario to a file stored on the instrument.

        Arguments:
            path: Full path to scenario file on the instrument.

        """

        # write folderpath as 'c:\folder\folder'
        writepath = self.fix_path_name(folderpath)

        self.backend.read(self.backend.inWaiting())
        self.backend.write(b'SAVE_SCENARIO,with_changes,as_simgen,%s/\n' % (writepath))
        self.current_scenario

    @staticmethod
    def fix_path_name(path):
        return path.replace('\\', '/')

    def write(self, key, returns=None):
        """Send a message to the spirent, and check the status message
        returned by the spirent.

        :return: Either 'value' (return the data response), 'status'
                 (return the instrument status), or None (raise an
                 exception if a data value is returned)
        """
        self.backend.read(self.backend.inWaiting())
        self._logger.debug('write {}'.format(repr(key)))
        self.backend.write('{}\n'.format(key))

        # Get the response
        response = b''
        while b'</msg>' not in response.lower():
            response += self.backend.readline()
        self._logger.debug(b'  <- {}'.format(repr(response)))
        self.backend.read(self.backend.inWaiting())  # Clear out any remaining data

        # Pull the data/error message payload
        data = re.match(rb'.*<data>[\W*]*(.*)[\W*]</data>.*', response, flags=re.S)

        if returns is None:
            if data is not None:
                raise Exception(data.group(1))
            return
        elif returns.lower() == b'value':
            return data.group(1)
        if returns.lower() == b'status':
            status = int(
                re.match(
                    rb'.*<status>[\W*]*(\d+)[\W*]</status>.*', response, flags=re.S
                ).group(1)
            )
            return status_messages[status]
        else:
            raise Exception(
                "Expected return type in ['value', 'status', None], but got {}".format(
                    repr(returns)
                )
            )

    def query(self, command):
        return self.write(command + '\n', returns='value')

    def run(self):
        """Start running the current scenario. Requires that there is time left in
        the scenario, otherwise run `rewind()` first.
        """
        self.write(b'RU')

    def end(self):
        """Stop running the current scenario. If a scenario is not
        running, an exception is raised.
        """
        self.write(b'-,EN')

    def rewind(self):
        """Rewind the current scenario to the beginning."""
        self.write(b'RW')

    def abort(self):
        """Force stop the current scenario."""
        self.write(b'-,EN,1,0')

    def reset(self):
        """End any currently running scenario, then rewind"""
        if self.status != b'ended':
            try:
                self.end()
            except BaseException:
                pass

        self.rewind()

    @attr.property.bytes(sets=False)
    def utc_time(self):
        """UTC time of the running scenario"""
        utc_unformatted = self.query(b'-,UTC_TIME')
        try:
            utc_struct = time.strptime(utc_unformatted, '%d-%b-%Y %H:%M:%S.%f')
            frac = utc_unformatted.split('.')[-1]
        except BaseException:
            utc_struct = time.strptime(utc_unformatted, '%d-%b-%Y %H:%M:%S')
            frac = b'000'

        return time.strftime('%Y-%m-%d %H:%M:%S', utc_struct) + '.' + frac

    @attr.property.bool(sets=False)
    def running(self):
        """`True` if a scenario is running."""
        return self.status == b'running'

    @attr.property.bytes(sets=False, only=status_messages, case=False)
    def status(self):
        """UTC time of the current running scenario."""
        return self.write(b'NULL', returns=b'status')


if __name__ == '__main__':
    lb.debug_to_screen('DEBUG')
    with SpirentGSS8000('COM17') as spirent:
        print(spirent.status)
        spirent.reset()
        spirent.run()
        scn = spirent.current_scenario
        utc = spirent.utc_time

    print(scn)
    print(utc)



"""
@authors: Dan Kuester <daniel.kuester@nist.gov>,
         Michael Voecks <michael.voecks@nist.gov>
"""

__all__ = [
    'LocalIPerf2',
    'LocalIPerf3',
    'AdbIPerf2',
    'LocalIPerf2Pair',
    'LocalPythonTrafficProfiler_ClosedLoopTCP',
]

import datetime
import re
import socket
import subprocess as sp
import time
import traceback
from io import StringIO
from pathlib import Path
from queue import Empty, Queue
from threading import Event, Thread
from time import perf_counter
from contextlib import AbstractContextManager, suppress
try:
    from ._networking import find_free_port
except ImportError as ex:
    if 'relative import' in str(ex):
        from _networking import find_free_port

import labbench as lb
from labbench import paramattr as attr
import ssmdevices.lib
import typing

if typing.TYPE_CHECKING:
    import pandas as pd
    import psutil
else:
    # delayed import for speed
    pd = lb.util.lazy_import('pandas')
    psutil = lb.util.lazy_import('psutil')

lb.util.force_full_traceback(True)
DataFrameType: typing.TypeAlias = 'pd.DataFrame'
SeriesType: typing.TypeAlias = 'pd.Series'

if __name__ == '__main__':
    from _networking import (
        get_ipv4_address,
        get_ipv4_occupied_ports,
        list_network_interfaces,
    )
else:
    from ._networking import (
        get_ipv4_address,
        get_ipv4_occupied_ports,
        list_network_interfaces,
    )

if '_tcp_port_offset' not in dir():
    _tcp_port_offset = 0

# Make sure the performance counter is initialized
perf_counter()


class IPerf2Values(lb.paramattr.HasParamAttrs):
    """base attributes shared by iperf2 and iperf3"""

    format: str = attr.value.str(
        default=None,
        only=('k', 'm', 'g', 'K', 'M', 'G'),
        help='data unit prefix in bits (k, m, g), bytes (K, M, G), or None for auto',
    )

    # address and network interface parameters
    client: str = attr.value.NetworkAddress(
        None,
        key='-c',
        accept_port=False,
        help='client host address (set None if server=True)',
    )
    server: bool = attr.value.bool(
        False,
        key='-s',
        help='True to run as a server'
    )
    port: int = attr.value.int(
        default=5201,
        key='-p',
        min=0,
        help='network port'
    )
    bind: str = attr.value.str(
        default=None,
        key='-B',
        help='bind connection to specified IP'
    )

    # timing and duration
    # (for time, default=None even though we know the default, because setting 10s conflicts with `number`)
    time: float = attr.value.float(
        None,
        key='-t',
        min=0,
        max=16535,
        allow_none=True,
        help='send duration (s) before quitting (default: 10)',
    )
    number: int = attr.value.int(
        None,
        key='-n',
        min=-1,
        allow_none=True,
        help='the number of bytes to transmit before quitting',
    )
    interval: float = attr.value.float(
        None,
        key='-i',
        min=0.01,
        label='s',
        help='seconds between throughput reports',
    )

    # high-level buffer commands
    udp: bool = attr.value.bool(
        False,
        key='-u',
        help='if True, use UDP instead of TCP'
    )
    bit_rate: str = attr.value.str(
        None,
        key='-b',
        label='bits/s',
        help='maximum bit rate, accepts KMG unit suffix; defaults 1Mbit/s UDP, no limit for TCP',
    )
    buffer_size: int = attr.value.int(
        None,
        key='-l',
        min=1,
        help='buffer size when generating traffic',
        label='bytes',
    )

    # TCP parameters
    tcp_window_size: int = attr.value.int(
        None,
        key='-w',
        min=1,
        allow_none=True,
        help='window / socket size (default is OS dependent?)',
        label='bytes',
    )
    nodelay: bool = attr.value.bool(
        False,
        key='-N',
        help='set True to use nodelay (TCP traffic only)'
    )
    mss: int = attr.value.int(
        None,
        key='-M',
        min=10,
        help='minimum segment size=MTU-40, TCP only',
        label='bytes',
    )

    def check_values(self):
        """validate configuration before a run"""

        # parameter conflict checks
        if self.client is not None and self.server:
            raise ValueError(
                'must set exactly one of (a) client operation by setting client=<client address>, '
                'or (b) server operation by setting server=True'
            )

        if self.udp:
            if self.mss is not None:
                raise ValueError('TCP MSS setting is incompatible with UDP')
            if self.nodelay:
                raise ValueError('TCP nodelay is incompatible with UDP')
            if self.buffer_size is not None:
                self._logger.warning(
                    'iperf may work improperly when setting udp=True with buffer_size'
                )

        if not self.udp and self.bit_rate is not None:
            raise ValueError('iperf does not support setting bit_rate in TCP')

        if self.server:
            if self.time is not None:
                raise ValueError('iperf server does not support the `time` argument')
            if self.number is not None:
                raise ValueError('iperf server does not support the `number` argument')

    def flags(self) -> list[str]:
        self.check_values()
        return lb.shell_options_from_keyed_values(self, hide_false=True)


class IPerf3Values(IPerf2Values):
    """command line argument values specific to iperf3"""

    reverse: bool = attr.value.bool(
        default=False, key='-R',
        help='run in reverse mode (server sends, client receives)',
    )
    json: bool = attr.value.bool(
        default=False, key='-J',
        help='output data in JSON format',
    )
    zerocopy: bool = attr.value.bool(
        default=False, key='-Z',
        help="whether to avoid buffer copies while sending data",
    )


class ShellIPerfBase(lb.ShellBackend):
    binary_name = attr.value.str(sets=False, cache=True, help='path (or name in system PATH) of the binary')
    binary_path = attr.value.str(None, cache=True, help='explicit path to the iperf binary (set automatically on open based on binary_name)')
    timeout = attr.value.float(5, cache=True, help='timeout waiting for output before an exception is raised')

    def open(self):
        if Path(self.binary_name).exists():
            self.binary_path = self.binary_name
        else:
            self.binary_path = ssmdevices.lib.path(self.binary_name, platform=True)

    def profile(self, *argv, block:bool=True):
        if not hasattr(self, 'time'):
            raise NotImplementedError('need to also inherit IPerf2Values or IPerf3Values to run the profiler')
        
        duration = 0 if self.time is None else self.time + 2
        timeout = max((self.timeout, duration))

        return super().run(
            *argv,
            *self.flags(),
            background=not block,
            pipe=True,
            respawn=not block,
            raise_on_stderr=True,
            timeout=timeout,
        )


class LocalIPerfBase(ShellIPerfBase):
    def profile(self, block: bool=True):
        self.check_ports()
        super().profile(self.binary_path, block=block)

    def check_ports(self):
        """check the availability of specified ports on the host"""
        if self.server:
            try:
                busy_ports = get_ipv4_occupied_ports(self.server)
            except psutil.AccessDenied:
                self._logger.warning('need administrator privileges on this platform to check for port access contention')
                busy_ports = []

            while self.port in busy_ports:
                prev_port = self.ports
                # find an open server port
                if self.port >= type(self).port.max:
                    self.port = type(self).port.min
                else:
                    self.port = self.port + 1
                self._logger.info(
                    f'requested port {prev_port} is in use - changing to {self.port}'
                )


class LocalIPerf3(LocalIPerfBase, IPerf3Values):
    """Run an instance of iperf3, collecting output data in a background thread.
    When running as an iperf client (server=False),
    The default value is the path that installs with 64-bit cygwin.
    """

    binary_name = attr.value.str('iperf3', inherit=True)

    # additional IPerf3-only
    reverse: bool = attr.value.bool(
        default=False, key='-R', help='run in reverse mode (server sends, client receives)'
    )
    json: bool = attr.value.bool(default=False, key='-J', help='output data in JSON format')
    zerocopy: bool = attr.value.bool(
        default=False, key='-Z', help="avoid buffer copies while sending data"
    )


class LocalIPerf2(LocalIPerfBase, IPerf2Values):
    """Run an instance of iperf to profile data transfer speed. It can
    operate as a server (listener) or client (sender), operating either
    in the foreground or as a background thread.
    When running as an iperf client (server=False).
    """

    DATAFRAME_COLUMNS = (
        'jitter_milliseconds',
        'datagrams_lost',
        'datagrams_sent',
        'datagrams_loss_percentage',
        'datagrams_out_of_order',
    )

    binary_name = attr.value.str('iperf', inherit=True)

    bidirectional: bool = attr.value.bool(
        default=False, key='-d', help='send and receive simultaneously'
    )
    report_style: str = attr.value.str(
        default='C',
        key='-y',
        only=('C', None),
        allow_none=True,
        help='"C" for DataFrame table output, None for formatted text',
    )

    def profile(self, block=True):
        ret = super().profile(block=block)
        if block:
            return self._format_output(ret)
        else:
            return ret

    def read_stdout(self):
        """retreive text from standard output. parse into a pandas DataFrame if self.report_style is None"""
        return self._format_output(super().read_stdout())

    def _format_output(self, stdout):
        """pack stdout into a pandas DataFrame if self.report_style == 'C'"""

        if self.report_style is None:
            return stdout.decode()

        # otherwise a dataframe
        columns = (
            'timestamp',
            'source_address',
            'source_port',
            'destination_address',
            'destination_port',
            'test_id',
            'interval',
            'transferred_bytes',
            'bits_per_second',
        )

        if self.udp:
            columns = columns + self.DATAFRAME_COLUMNS

        if isinstance(stdout, bytes):
            stdout_buf = StringIO(stdout.decode())
        else:
            stdout_buf = StringIO(stdout)

        data = pd.read_csv(stdout_buf, header=None, index_col=False, names=columns)

        if data.shape[1] > 0:
            data.drop(
                ['interval', 'transferred_bytes', 'test_id'],
                inplace=True, axis=1
            )
        data['timestamp'] = pd.to_datetime(data['timestamp'], format='%Y%m%d%H%M%S')
        frac_sec = (data.index * self.interval) % 1
        data['timestamp'] = data['timestamp'] + pd.TimedeltaIndex(frac_sec, 's')

        return data


class AdbIPerf2(ShellIPerfBase, IPerf2Values):
    # leave this as a string to avoid validation pitfalls if the host isn't POSIXey
    binary_name = attr.value.str('adb', inherit=True)
    remote_binary_path = attr.value.str('/data/local/tmp/iperf', cache=True, help='copy destination for iperf in the handset')

    def profile(self, block=True):
        ret = super().profile(self.binary_path, 'shell', self.remote_binary_path, block=block)

        if block:
            return self._format_output(ret)
        else:
            # wait for output before returning
            test = self.read_stdout(1)
            if 'network' in test:
                self._logger.warning('no network connectivity in UE')

            return ret

    def open(self):
        self.wait_for_device(30)

        self._logger.debug('copying iperf onto phone')
        lb.sleep(0.1)
        local_iperf_path = ssmdevices.lib.path('iperf', platform='android-amd64')
        self.run(self.binary_path, 'push', local_iperf_path, self.remote_binary_path)
        self.wait_for_device(2)

        # set permissions for execution
        self.run(self.binary_path, 'shell', 'chmod', '777', self.remote_binary_path, check=False)
        self.wait_for_device(2)

        # validate
        stdout = self.run(self.binary_path, 'shell', self.remote_binary_path, '--help', timeout=2, pipe=True)
        if stdout.startswith(b'/system/bin/sh'):
            # adb dumps both stderr and stdout from the handset into stdout, so we get little
            # from monitoring. if iperf ran correctly, however, there is no message from sh
            raise OSError(f'adb shell iperf --help failed: {stdout}')

        self._logger.debug('phone is ready to execute iperf')

    def kill(self, wait_time=3):
        """Kill the local process and the iperf process on the UE."""

        if self.binary_path is None:
            return

        # Find and kill processes on the UE
        out = self.run(self.binary_path, 'shell', 'ps')
        for line in out.splitlines():
            line = line.decode(errors='replace')
            if self.remote_binary_path in line.lower():
                pid = line.split()[1]
                stdout = self.pipe('shell', 'kill', '-9', pid)
                self._logger.debug(f'killing zombie iperf: {stdout}')

            lb.sleep(0.1)
            # Wait for any iperf zombie processes to die
            t0 = time.time()
            while time.time() - t0 < wait_time and wait_time != 0:
                out = self.pipe('shell', 'ps').lower()
                if b'iperf' not in out:
                    break
                lb.sleep(0.25)
            else:
                raise TimeoutError(
                    'timeout waiting for iperf process termination on UE'
                )

        # Kill the local adb process
        super().kill()

    def read_stdout(self):
        """adb seems to forward stderr as stdout. Filter out some undesired
        resulting status messages.
        """
        txt = lb.ShellBackend.read_stdout(self)

        # remove extra output added by adb
        out = []
        for line in txt.splitlines():
            if b':' not in line:
                out.append(line)
            else:
                self._logger.warning('stdout: {}'.format(repr(line)))
        out = b'\n'.join(out)

        return self._format_output(out)

    def wait_for_cell_data(self, timeout: float=60):
        """block until cellular data is available

        Arguments:
            timeout: wait time in seconds before raising TimeoutError
        """

        self._logger.debug('waiting for cellular data connection')
        t0 = time.perf_counter()
        out = ''

        while time.perf_counter() - t0 < timeout or timeout is None:
            out = sp.run(
                [self.binary_path, 'shell', 'dumpsys', 'telephony.registry'],
                stdout=sp.PIPE,
                check=True,
                timeout=timeout,
            ).stdout

            con = re.findall(
                r'mDataConnectionState=([\-0-9]+)', out.decode(errors='replace')
            )

            if len(con) > 0 and con[0] == '2':
                break
        else:
            raise TimeoutError('phone did not connect for cellular data before timeout')

        self._logger.debug(f'cellular data available after {time.perf_counter() - t0} s')

    def reboot(self, block=True):
        """reboot the handset.

        Arguments:
            block: if truey, block until the device is again ready to accept commands
        """
        self._logger.info('rebooting')
        self.run(self.binary_path, 'reboot')
        if block:
            self.wait_for_device()

    def wait_for_device(self, timeout=30):
        """block until the device is ready to accept commands"""

        self._logger.debug('awaiting USB connection to handset')
        self.run(self.binary_path, 'wait-for-device', timeout=timeout)


class LocalIPerf2Pair(LocalIPerf2):
    """Configure and run an iperf client and a server pair on the host.

    The socket connection between the two is bound to ensure routing across
    the two specified network interfaces (instead of e.g., loopback).
    """

    # add other settings
    server: str = attr.value.NetworkAddress(
        None, accept_port=False, help='the ip address where the server listens'
    )
    client: str = attr.value.NetworkAddress(
        None, accept_port=False, help='the ip address from which the client sends data'
    )

    children = {}

    def open(self):
        self.children: dict[str, LocalIPerf2] = dict()
        self.backend: None = None

    def close(self):
        try:
            self.kill()
        except TypeError as e:
            if 'NoneType' not in str(e):
                raise

    def kill(self):
        for name in ('client', 'server'):
            child = self.children.pop(name, None)
            if child is not None:
                child.kill()

    def running(self):
        for name in ('client', 'server'):
            child = self.children.get(name, None)

            if child is None:
                continue

            if child.running():
                return True

        return False

    def profile(self, block=True, **kws):
        if not block:
            kws['time'] = 0
            kws['number'] = -1
        self._setup_pair(**kws)

        self.children['server'].profile(block=False)
        ret = self.children['client'].profile(block=block)

        if block:
            ret = self.read_stdout(client_ret=ret)
            self.kill()
            return ret

    def read_stdout(self, client_ret: typing.Union[str, None]=None) -> typing.Union[dict, DataFrameType]:
        try:
            if client_ret is None:
                client = self.children['client']
            server = self.children['server']
        except KeyError:
            return dict(client=None, server=None)
        
        if client_ret is None:
            client_stdout = client.read_stdout()
        else:
            client_stdout = client_ret
        server_stdout = server.read_stdout()

        if isinstance(client_stdout, pd.DataFrame):
            # a single merged DataFrame
            return self._merge_dataframes(client_stdout, server_stdout)
        else:
            # a dictionary of text results
            return dict(client=client_stdout, server=server_stdout)

    def _merge_dataframes(self, client, server):
        client.columns = [
            ('client_' if n != 'timestamp' else '') + str(n) for n in client.columns
        ]

        server.columns = [
            ('server_' if n != 'timestamp' else '') + str(n) for n in server.columns
        ]

        return client.merge(server, how='outer', on='timestamp')

    def _setup_pair(self, **kws):
        if self.running():
            raise BlockingIOError(f'{self} is already running')

        # override parameters as necessary for each side of the link
        client = self.children['client'] = LocalIPerf2()
        server = self.children['server'] = LocalIPerf2()

        # seed defaults from self
        for name, attr_def in lb.paramattr.get_class_attrs(self).items():
            if isinstance(attr_def, lb.paramattr.value.Value) and attr_def.sets:
                value = getattr(self, name)
            else:
                continue

            if value != attr_def.default:
                setattr(client, name, value)
                setattr(server, name, value)

        # override with client/server specifics
        client.client=self.client
        client.bind=f'{self.client}:{find_free_port()}'
        client.server=False
        client.check_ports()

        server.client=None
        server.bind=self.server
        server.server=True
        server.time=None
        server.number=None
        server.check_ports()

        self.backend = lb.sequentially(server, client).__enter__()


m1 = 0x5555555555555555
m2 = 0x3333333333333333
m4 = 0x0F0F0F0F0F0F0F0F
m8 = 0x00FF00FF00FF00FF
m16 = 0x0000FFFF0000FFFF
m32 = 0x00000000FFFFFFFF
h01 = 0x0101010101010101


def bit_errors(x):
    """See: https://en.wikipedia.org/wiki/Hamming_weight"""
    import numpy as np

    if x is None:
        return None
    #    a1 = np.frombuffer(buf1,dtype='uint64')
    x = np.frombuffer(x[: (len(x) // 8) * 8], dtype='uint64').copy()
    x -= (x >> 1) & m1
    x = (x & m2) + ((x >> 2) & m2)
    x = (x + (x >> 4)) & m4
    return ((x * h01) >> 56).sum()


class LocalPythonTrafficProfiler(lb.Device):
    """Profile closed-loop traffic between two network interfaces
    on this computer. Takes advantage of the system clock as a common
    basis for traffic delay measurement, with uncertainty on the scale
    of the the system clock tick resolution.
    """

    server: str = attr.value.str(
        None, help='the name of the network interface that will send data'
    )
    client: str = attr.value.str(
        None, help='the name of the network interface that will receive data'
    )
    receive_side: str = attr.value.str(
        None,
        only=('server', 'client'),
        help='which of the server or the client does the receiving',
    )
    port: int = attr.value.int(
        0,
        min=0,
        help='TCP or UDP port for networking, or 0 to let the operating system choose',
    )
    timeout: float = attr.value.float(
        2, min=1e-3, help='timeout before aborting the test', cache=True
    )
    tcp_nodelay: bool = attr.value.bool(
        True, help="set True to disable Nagle's algorithm"
    )
    sync_each: bool = attr.value.bool(
        False,
        help='synchronize the start times of the send and receive threads for each buffer at the cost of throughput',
    )

    delay: float = attr.value.float(
        0, min=0, help='wait time before profiling', cache=True
    )

    def __repr__(self):
        return "{name}(server='{server}',client='{client}')".format(
            name=self.__class__.__name__, server=self.server, client=self.client
        )

    def close(self):
        if self.is_running():
            self.stop_traffic()

    def start(self, buffer_size, count=None, duration=None):
        """Start a background thread that runs a one-way traffic test.

        It will end when `count` buffers have been tested, `duration`
        time has elapsed, or `stop_traffic` is called. To retrieve the
        traffic data, call `stop_traffic`.
        """
        self._background_event = Event()
        self._background_queue = Queue()

        server_sock, client_sock, listener = self._open_sockets(buffer_size)

        try:
            self._run(
                client_sock=client_sock,
                server_sock=server_sock,
                buffer_size=buffer_size,
                end_event=self._background_event,
                count=count,
                duration=duration,
            )
        except:
            self._close_sockets(client_sock, server_sock, listener)
            raise

    def is_running(self):
        return (
            hasattr(self, '_background_event') and not self._background_event.is_set()
        )

    def get(self):
        if not hasattr(self, '_background_queue'):
            raise ChildProcessError('no traffic history, start a run first')

        try:
            ret = self._background_queue.get(timeout=self.timeout)
        except Empty:
            ret = {}

        if isinstance(ret, BaseException):
            raise ret
        else:
            return self._make_dataframe(
                ret,
            )

    def stop(self):
        if not hasattr(self, '_background_queue'):
            raise ChildProcessError('no traffic running, start a run first')

        self._background_event.set()
        return self.get()

    def _make_dataframe(self, data):
        raise NotImplementedError


class suppress_matching_arg0(AbstractContextManager):
    """Context manager to suppress specified exceptions that must also match
        a specified first argument.

    After the exception is suppressed, execution proceeds with the next
    statement following the with statement.

         with suppress(FileNotFoundError):
             os.remove(somefile)
         # Execution still resumes here if the file was already removed
    """

    def __init__(self, *exceptions, arg0=None):
        self._exceptions = exceptions
        self._arg0 = arg0

    def __enter__(self):
        pass

    def __exit__(self, exctype, excinst, exctb):
        if exctype is None:
            return False

        if not issubclass(exctype, self._exceptions):
            return True

        if self._arg0 is None or self._arg0 == excinst.args[0]:
            return False
        else:
            return True


class ServerConnectionError(ConnectionError):
    pass


class ClientConnectionError(ConnectionError):
    pass


class PortBusyError(ConnectionError):
    pass


class LocalPythonTrafficProfiler_ClosedLoopTCP(LocalPythonTrafficProfiler):
    _server = None
    PORT_WINERRS = (10013, 10048)
    CONN_WINERRS = (10051,)

    def _close_sockets(self, *sockets, bytes_=0):
        for sock in sockets:
            try:
                sock.settimeout(0.1)
            except OSError:
                continue

            with suppress_matching_arg0(OSError, arg0=10057):
                sock.send(b'')

            with suppress(OSError):
                t0 = perf_counter()
                while perf_counter() - t0 < 1:
                    try:
                        buf = sock.recv(bytes_)
                    except socket.timeout:
                        break
                    if len(buf) == 0:
                        break
                else:
                    self._logger.warning('failed to flush socket before closing')

            with (
                suppress_matching_arg0(OSError, arg0=10057),
                suppress_matching_arg0(OSError, arg0='timed out'),
            ):
                sock.shutdown(socket.SHUT_RDWR)

            with suppress_matching_arg0(OSError, arg0=10057):
                sock.close()

    @property
    def _receive_interface(self):
        if self.receive_side == 'server':
            return self.server
        else:
            return self.client

    def _open_sockets(self, buffer_size):
        """Connect the supplied client socket to the server."""

        global _tcp_port_offset

        server_ip = get_ipv4_address(self.server)
        client_ip = get_ipv4_address(self.client)

        timeout = self.timeout
        bytes_ = buffer_size
        tcp_nodelay = self.tcp_nodelay

        client_done = Event()
        server_done = Event()

        def listener(port):
            """Run a listener at the socket"""

            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(timeout)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

                # Set the size of the buffer for this socket
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, bytes_)
                self._logger.info(
                    f'created server socket with recv buffer size {bytes_}'
                )
                bufsize = sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
                if bufsize < bytes_:
                    msg = (
                        f'recv buffer size is {bufsize}, but need at least {self.bytes}'
                    )
                    raise OSError(msg)
                self._logger.info(f'binding listener to {server_ip}:{port}')
                sock.bind((server_ip, port))

                # start listening
                sock.listen(5)
            except OSError as e:
                if hasattr(e, 'winerror') and e.winerror in self.PORT_WINERRS:
                    raise PortBusyError()
                else:
                    raise

            return sock

        def client(listen_sock):
            port = listen_sock.getsockname()[1]
            sock = None

            self._logger.warning('client start')

            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(timeout)

                # Basic flags
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, tcp_nodelay)

                # Set and verify the send buffer size
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, bytes_)
                bytes_actual = sock.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
                if bytes_actual != bytes_:
                    msg = f'client buffer size is {bytes_actual}, but requested {self.bytes}'
                    raise OSError(msg)

                try:
                    # binding to the specific port creates a conflict
                    # with the server, and a connection failure
                    sock.bind((client_ip, 0))
                except OSError as e:
                    ex = e
                else:
                    ex = None

                if ex is not None:
                    raise ConnectionRefusedError(*ex.args)

                # Do the connect
                sock.connect((server_ip, port))

            # This exception needs to come first, because it is a subclass
            # of OSError (at least on windows)
            except socket.timeout:
                if sock is not None:
                    self._close_sockets(sock, bytes_=bytes_)
                msg = f'client socket timed out in attempt to connect to the server at {server_ip}:{port}'
                ex = ConnectionRefusedError(msg)

                # Certain "port busy" errors are raised as OSError. Check whether
            # they match a whitelist of known port errors to map into port busy
            except OSError as e:
                msg = f'connection failed between server {server_ip} and client {client_ip}'
                # Windows-specific errors
                if hasattr(e, 'winerror') and e.winerror in self.PORT_WINERRS:
                    self._logger.debug(msg)
                    ex = PortBusyError(msg)
                elif hasattr(e, 'winerror') and e.winerror in self.CONN_WINERRS:
                    self._logger.debug(msg)
                    ex = ConnectionError(msg)
                else:
                    raise

            # For everything else, we still need to clean up
            except BaseException:
                if sock is not None:
                    self._close_sockets(sock, bytes_=bytes_)
                raise
            else:
                ex = None

            if ex is not None:
                raise ex

            # if this times out, some kind of exception should have been
            # raised in the server thread. leave quietly and allow the
            # exception to come from the server.
            client_done.set()
            if not server_done.wait(timeout):
                self._close_sockets(sock, bytes_=bytes_)

            return sock

        def server(listen_sock):
            self._logger.warning('server start')

            conn = None
            ex = None
            # Try to get a connection
            t0 = perf_counter()

            try:
                while perf_counter() - t0 < timeout:
                    try:
                        listen_sock.settimeout(0.051 + timeout - (perf_counter() - t0))
                        self._sock = listen_sock
                        conn, (other_ip, _) = listen_sock.accept()
                    except socket.timeout:
                        continue

                    if other_ip == client_ip:
                        break
                    else:
                        port = listen_sock.getsockname()[1]
                        self._logger.warning(
                            f'connection attempt from unexpected ip {other_ip} instead of {client_ip}:{port}'
                        )
                        if conn is not None:
                            self._close_sockets(conn, bytes_=bytes_)
                            conn = None
                else:
                    port = listen_sock.getsockname()[1]
                    raise TimeoutError(
                        f'socket server received no connection attempt on {server_ip}:{port}'
                    )
            except BaseException as e:
                listen_sock.settimeout(timeout)
                ex = e
            else:
                listen_sock.settimeout(timeout)

                # Basic flags
                conn.settimeout(timeout)
                #            conn.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, tcp_nodelay)

                # Set and verify the send buffer size
                conn.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, bytes_)
                bytes_actual = conn.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
                if bytes_actual != bytes_:
                    msg = f'server buffer size is {bytes_actual}, but requested {self.bytes}'
                    raise OSError(msg)

            if not client_done.wait(timeout):
                # Suppress the server exception if the client is already
                # raising one
                if ex is not None:
                    self._logger.debug(
                        f'server connection exception: {repr(ex)} (superceded by client exception)'
                    )
                    ex = None
                if conn is not None:
                    self._close_sockets(conn, bytes_=bytes_)

            server_done.set()

            if ex is not None:
                raise ex

            return conn

        def open_():
            global _tcp_port_offset, listen_sock

            if self.port != 0:
                port = self.port + _tcp_port_offset
            else:
                port = 0

            listen_sock = listener(port)

            # Keep trying on many ports
            try:
                ret = lb.concurrently(
                    lb.Call(server, listen_sock=listen_sock),
                    lb.Call(client, listen_sock=listen_sock),
                )
            except:
                self._close_sockets(listen_sock, bytes_=0)
                raise
            finally:
                if self.port != 0:
                    _tcp_port_offset = (_tcp_port_offset + 1) % 5000

            ret['listener'] = listen_sock
            return ret

        try:
            t0 = perf_counter()

            if self.port == 0:
                ret = open_()
            else:
                # Allow chances to try other ports
                ret = lb.retry(PortBusyError, 100)(open_)()
            p = ret['client'].getsockname()[1]
            self._logger.debug(
                f'server {server_ip}:{p} accepted connection from client {client_ip}:{p} in {perf_counter() - t0:0.3f}s'
            )
        except PortBusyError:
            raise ConnectionError(r'failed to connect on {retries} ports')

        return ret['server'], ret['client'], ret['listener']

    def _run(
        self,
        client_sock,
        server_sock,
        buffer_size,
        duration=None,
        count=None,
        end_event=None,
    ):
        if duration is count is end_event is None:
            raise ValueError(
                'must pass at least one of duration, count, and end_event to specify end condition'
            )

        if self.tcp_nodelay and buffer_size < self.mss():
            raise ValueError(
                f'with tcp_nodelay enabled, set buffer_size at least as large as the MSS ({self.mss()})'
            )

        if self.server == self._receive_interface:
            send_sock, recv_sock = client_sock, server_sock
        else:
            send_sock, recv_sock = server_sock, client_sock

        t_start = perf_counter()

        # Are we running in the background?
        background = end_event is not None

        if background:
            count = 1

        # Pull some parameters and thread sync objects into the namespace
        timeout = self.timeout
        bytes_ = buffer_size
        sync = self.sync_each
        delay = self.delay
        rx_ready = Event()
        tx_ready = Event()

        if end_event is None:
            end_event = Event()

        except_event = Event()

        def check_status():
            lb.sleep(0)

            # Bail if the sender hit an exception
            if except_event.is_set():
                raise lb.util.ThreadEndedByMaster()

        def traffic_done(i):
            if background and end_event.is_set():
                return True
            elif (count is not None) and i >= count:
                return True
            elif (duration is not None) and perf_counter() - t_start >= duration:
                return True
            else:
                return False

        def receiver():
            buf = bytearray(bytes_)
            starts = []
            finishes = []
            rx_buffer0_finishes = []
            rx_buffer_sizes = []
            rx_buffers = []

            def do_sync():
                check_status()
                if sync:
                    t_sync = perf_counter()
                    rx_ready.set()
                    while not tx_ready.is_set():
                        if perf_counter() - t_sync > timeout:
                            except_event.set()
                            raise TimeoutError('timeout waiting for sender sync')
                    tx_ready.clear()

            def single():
                """Receive a single buffer of data"""
                bytes_left = int(bytes_)
                do_sync()
                t0 = t1 = t2 = perf_counter()
                i = 0

                # Receive rx_buffers until we receive the full send buffer
                while bytes_left > 0:
                    if t2 - t0 > timeout:
                        msg = f'timeout while waiting to receive {bytes_left} of {bytes_} bytes'
                        raise TimeoutError(msg)
                    try:
                        bytes_left -= recv_sock.recv_into(buf[-bytes_left:], bytes_left)
                        if i == 0:
                            rx_buffer0_size = int(bytes_) - bytes_left
                            t1 = t2 = perf_counter()
                        else:
                            t2 = perf_counter()
                        i += 1
                    except socket.timeout as e:
                        raise TimeoutError(' '.join(e.args))

                return t0, t1, t2, rx_buffer0_size, i

            try:
                do_sync()

                # One to the wind
                single()
                rx_ready.set()

                i = 0
                # Receive the test data
                while not traffic_done(i):
                    t0, t1, t2, rx_buffer0_size, rx_buffer_count = single()
                    starts.append(t0)
                    rx_buffer0_finishes.append(t1)
                    finishes.append(t2)
                    rx_buffer_sizes.append(rx_buffer0_size)
                    rx_buffers.append(rx_buffer_count)

                    i += 1

                # One to the wind
            #                single()

            except lb.util.ThreadEndedByMaster:
                self._logger.debug(
                    f'{self.__class__.__name__}() ended by master thread'
                )
                except_event.set()
            except BaseException:
                if not (end_event is not None and end_event.is_set()):
                    except_event.set()
                    raise

            return {
                't_rx_start': starts,
                't_rx_end_buffer0': rx_buffer0_finishes,
                't_rx_end': finishes,
                'rx_buffer0_size': rx_buffer_sizes,
                'rx_buffer_count': rx_buffers,
            }

        def sender():
            """This runs in the receive thread, with the socket connected."""
            start_timestamps = []
            finish_timestamps = []
            start = None

            def do_sync():
                check_status()
                if sync:
                    t_sync = perf_counter()
                    while not rx_ready.is_set():
                        if perf_counter() - t_sync > timeout:
                            except_event.set()
                            raise TimeoutError('timeout waiting for receive sync')
                    rx_ready.clear()
                    tx_ready.set()

            def single():
                import numpy as np

                data = np.random.bytes(bytes_)
                do_sync()
                t0 = t1 = perf_counter()
                try:
                    send_sock.sendall(data)
                    t1 = perf_counter()
                    if delay > 0:
                        time.sleep(delay)
                except socket.timeout:
                    ex = IOError('timed out attempting to send data')
                else:
                    ex = None

                return t0, t1, ex

            try:
                do_sync()

                # Throwaway buffer
                single()
                tx_ready.set()

                i = 0
                while not traffic_done(i):
                    if i == 0:
                        start = datetime.datetime.now()

                    t0, t1, ex = single()
                    if ex is not None:
                        raise ex

                    start_timestamps.append(t0)
                    finish_timestamps.append(t1)

                    i += 1

                # Throwaway buffer
            #                single()

            except lb.util.ThreadEndedByMaster:
                self._logger.debug(
                    f'{self.__class__.__name__}() ended by master thread'
                )
                except_event.set()
            except BaseException as e:
                if not (end_event is not None and end_event.is_set()):
                    self._logger.debug(f'suppressed exception in sender: {e}')
                    except_event.set()

            return {
                'start': start,
                'bytes': bytes_,
                't_tx_start': start_timestamps,
                't_tx_end': finish_timestamps,
            }

        def background_thread():
            try:
                ret = None
                i = 0
                while not end_event.is_set():
                    ret = lb.concurrently(sender, receiver, traceback_delay=True)
                    self._background_queue.put(ret)
                    i += 1
                self._logger.debug(
                    f'finished traffic test of {i} buffers of {bytes_} bytes'
                )
            except BaseException as e:
                if not self._background_event.is_set():
                    self._logger.warning(
                        f'background thread exception - traceback: {traceback.format_exc()}'
                    )
                    self._background_queue.put(e)
                self._close_sockets(
                    send_sock, recv_sock, listen_sock, bytes_=buffer_size
                )
            finally:
                self._logger.debug('background thread finished')

        if background:
            thread = Thread(target=background_thread)
            thread.start()
            tx_ready.wait_for_device(timeout=self.timeout)
            rx_ready.wait_for_device(timeout=self.timeout)
            self._logger.debug(
                f'first buffer sent after {perf_counter() - t_start:0.3f}s'
            )
        else:
            ret = lb.concurrently(sender, receiver, traceback_delay=True)
            i = len(ret['t_tx_start'])
            self._logger.debug(
                f'finished traffic test of {i} buffers of {bytes_} bytes'
            )
            return ret

    def profile_count(self, buffer_size: int, count: int):
        """sends `count` buffers of size `buffer_size` bytes
        and returns profiling information"

        Arguments:

            buffer_size (int): number of bytes to send in each buffer

            count (int): the number of buffers to send

        :returns: a DataFrame indexed on PC time containing columns 'bits_per_second', 'duration', 'delay', 'queuing_duration'
        """

        # Connect all the sockets
        server_sock, client_sock, listener = self._open_sockets(buffer_size)

        try:
            ret = self._run(
                client_sock=client_sock,
                server_sock=server_sock,
                count=count,
                buffer_size=buffer_size,
            )

        finally:
            self._close_sockets(client_sock, server_sock, listener)

        return self._make_dataframe(ret)

    def profile_duration(self, buffer_size: int, duration: float):
        """sends buffers of size `buffer_size` bytes until
        `duration` seconds have elapsed, and returns profiling information"

        Arguments:

            buffer_size (int): number of bytes to send in each buffer

            duration (float): the minimum number of seconds to spend profiling

        :returns: a DataFrame indexed on PC time containing columns 'bits_per_second', 'duration', 'delay', 'queuing_duration'
        """
        #        t0 = perf_counter()

        # Connect all the sockets
        server_sock, client_sock, listener = self._open_sockets(buffer_size)

        try:
            ret = self._run(
                client_sock=client_sock,
                server_sock=server_sock,
                buffer_size=buffer_size,
                duration=duration,
            )

        finally:
            self._close_sockets(client_sock, server_sock, listener)

        return self._make_dataframe(ret)

    def _make_dataframe(self, worker_data):
        self._logger.debug('making dataframe')
        start = worker_data.pop('start', None)
        if start is None:
            raise IOError('the run did not return data')

        buffer_size = worker_data.pop('bytes')

        # Race condition may make the lengths different by 1
        count = min(len(worker_data['t_tx_start']), len(worker_data['t_rx_start']))

        for k in worker_data.keys():
            worker_data[k] = worker_data[k][:count]

        # Compute elapsed timings, shift timestamps, etc.
        ret = pd.DataFrame(worker_data)

        timestamp = pd.TimedeltaIndex(ret.t_tx_start, unit='s') + start

        duration = ret.t_rx_end - ret.t_rx_start

        # The average data rate after the first partial receive buffer
        late_rate = (buffer_size - ret.rx_buffer0_size) / (
            ret.t_rx_end - ret.t_rx_end_buffer0
        )

        # Estimate the clock value immediately before the data arrived at the
        # receive socket based on the remainder of the data
        est_rx_buffer0_start = ret.t_rx_end_buffer0 - ret.rx_buffer0_size / late_rate

        ret = pd.DataFrame(
            {
                'bits_per_second': 8 * buffer_size / duration,
                'duration': duration,
                'delay': est_rx_buffer0_start
                - ret.t_tx_start,  # ret.t_rx_start-ret.t_tx_start,
                'queuing_duration': ret.t_tx_end - ret.t_tx_start,
                'rx_buffer_count': ret.rx_buffer_count,
                't_rx_end_buffer0': ret.t_rx_end_buffer0,
                'timestamp': timestamp,
            }
        )

        return ret.set_index('timestamp')

    def mss(self):
        return self.mtu() - 40

    def mtu(self):
        iface = list_network_interfaces('interface')[self._receive_interface][
            'interface'
        ]
        return psutil.net_if_stats()[iface].mtu

    def wait_for_interfaces(self, timeout):
        errors = (TimeoutError, ConnectionRefusedError)

        t0 = time.perf_counter()
        socks = lb.until_timeout(errors, timeout)(self._open_sockets)()
        self._close_sockets(*socks)
        elapsed = perf_counter() - t0
        self._logger.debug(f'interfaces ready after {elapsed:0.2f}s')
        return elapsed

def test_iperf2_bound_pair_blocking():
    # When both network interfaces run on the same computer,
    # it is convenient to use IPerf2BoundPair, which runs both
    # an iperf server and an iperf client. each socket is bound
    # to these interfaces to ensure that traffic is routed through
    # the devices under test.
    iperf = LocalIPerf2Pair(
        server='127.0.0.1',
        client='127.0.0.1',

        ### the parameters below set the corresponding iperf command line flags
        # tcp_window_size=8196# -w (default unknown?)
        # buffer_size='16k'   # -l (default unknown? possible strange results for UDP)
        interval=1,         # -i (default is no output until the end of process)
        # bidirectional=True, # -d (default is False)
        udp=True,           # -u (default is False (TCP))
        # bit_rate='1M'       # -b (default is no bit rate throttling)
        time=10,            # -t (how long to run the iperf client; iperf's default is 10s)
        # report_style='C',   # -y (we set this from python to 'C' by default for CSV table; set to None for text output)
        # number=-1,        # -n (by default, iperf uses -t to determine client test length; set -1 to run until killed)
        # nodelay=True,       # -N (default is False; TCP only)
        # mss=1460,           # -M (default 1460? - TCP only, of course)
    )

    with iperf:
        iperf.time = 3
        data = iperf.profile(block=True)
        print(data)


def test_iperf2_bound_pair_background():
    # When both network interfaces run on the same computer,
    # it is convenient to use IPerf2BoundPair, which runs both
    # an iperf server and an iperf client. each socket is bound
    # to these interfaces to ensure that traffic is routed through
    # the devices under test.
    iperf = LocalIPerf2Pair(
        server='127.0.0.1',
        client='127.0.0.1',

        ### the parameters below set the corresponding iperf command line flags
        # tcp_window_size=8196# -w (default unknown?)
        # buffer_size='16k'   # -l (default unknown? possible strange results for UDP)
        interval=1,         # -i (default is no output until the end of process)
        # bidirectional=True, # -d (default is False)
        udp=True,           # -u (default is False (TCP))
        # bit_rate='1M'       # -b (default is no bit rate throttling)
        time=10,            # -t (how long to run the iperf client; iperf's default is 10s)
        report_style='C',   # -y (we set this from python to 'C' by default for CSV table; set to None for text output)
        # number=-1,        # -n (by default, iperf uses -t to determine client test length; set -1 to run until killed)
        # nodelay=True,       # -N (default is False; TCP only)
        # mss=1460,           # -M (default 1460? - TCP only, of course)
    )

    # profile returns immediately when block=False, leaving iperf running in the background.
    # this allows other tasks here in the main thread
    with iperf:
        iperf.profile(block=False)
        time.sleep(4) # replace this with other code for automating other equipment
        data = iperf.read_stdout()
        iperf.kill()

        # data is returned as a pandas dataframe.
        data.to_csv(r'c:\users\dkuester\output.csv')

        assert len(data) > 0


def test_local_python_closed_loop():
    lb.show_messages('debug')

    # try the first interface available on the system
    iface = tuple(list_network_interfaces('interface').keys())[0]

    net = LocalPythonTrafficProfiler_ClosedLoopTCP(
        server=iface,
        client=iface,
        receive_side='server',
        port=0,
        tcp_nodelay=True,
        timeout=2,
    )

    with net:
        ret = net.profile_count(10 * net.mtu(), count=10)
        assert len(ret) > 0


def test_separate_iperf2():
    lb.show_messages('debug')

    server = LocalIPerf2(server=True, port=5050, interval=0.25, udp=True)
    client = LocalIPerf2(client="127.0.0.1", port=5054, time=10, interval=0.25)

    with server, client:
        server.profile(block=False)

        client.open()
        client.profile(block=False)
        time.sleep(5)
        client.kill()
        client_out = client.read_stdout()
        assert len(client_out) > 0


# -*- coding: utf-8 -*-

__all__ = [
    'KeysightU2000XSeries',
    'KeysightU2044XA',
    'RohdeSchwarzNRP8s',
    'RohdeSchwarzNRP18s',
    'PowerTrace_RohdeSchwarzNRP',
]

import labbench as lb
from labbench import paramattr as attr
import typing
import typing_extensions
import warnings
import contextlib
import time
import pyvisa

if typing.TYPE_CHECKING:
    import pandas as pd
    import numpy as np
else:
    # delayed import for speed
    pd = lb.util.lazy_import('pandas')
    np = lb.util.lazy_import('numpy')

DataFrameType: typing_extensions.TypeAlias = 'pd.DataFrame'
SeriesType: typing_extensions.TypeAlias = 'pd.Series'

class KeysightU2000XSeries(lb.VISADevice):
    """Coaxial power sensors connected by USB"""

    # used for automatic connection
    make = attr.value.str('Keysight Technologies', inherit=True)

    def open(self):
        if type(self) is KeysightU2000XSeries:
            warnings.warn(
                'the generic base class KeysightU2000XSeries will be removed from the '
                'ssmdevices.instruments namespace in a future release. instead, use '
                'a specific model, such as KeysightU2044XA',
                DeprecationWarning,
            )

        self._clear()
        self._event_status_enable()

    initiate_continuous = attr.property.bool(
        key='INIT:CONT', help='whether to enable triggering to acquire power samples'
    )
    output_trigger = attr.property.bool(
        key='OUTP:TRIG', help='whether to enable the trigger output signal'
    )
    trigger_source = attr.property.str(
        key='TRIG:SOUR',
        case=False,
        only=('IMM', 'INT', 'EXT', 'BUS', 'INT1'),
        help='input source for the acquisitiont trigger',
    )
    trigger_count = attr.property.int(
        key='TRIG:COUN',
        min=1,
        max=200,
        help='the number of contiguous power samples to acquire per trigger',
    )
    measurement_rate = attr.property.str(
        key='SENS:MRAT',
        only=('NORM', 'DOUB', 'FAST'),
        case=False,
        help='measurement speed',
    )
    sweep_aperture = attr.property.float(
        key='SWE:APER',
        min=20e-6,
        max=200e-3,
        help='duration of the acquisition window of each power sample',
        label='s',
    )
    frequency = attr.property.float(
        key='SENS:FREQ',
        min=10e6,
        max=18e9,
        step=1e-3,
        label='Hz',
        help='input signal center frequency',
    )
    auto_calibration = attr.property.bool(
        key='CAL:ZERO:AUTO', help='whether to enable auto-calibration'
    )
    options = attr.property.str(
        key='*OPT',
        sets=False,
        cache=True,
        help='a comma-separated list of installed license options',
    )

    def preset(self, wait=True) -> None:
        """restore the instrument to its default state"""
        self.write('SYST:PRES')
        if wait:
            self.wait()
        self._clear()
        self._event_status_enable()

    def fetch(self, precheck=True) -> typing.Union[float, SeriesType]:
        """return power readings from the instrument.

        Returns:
            a single number if trigger_count == 1, otherwise or pandas.Series"""
        
        if precheck:
            self._check_errors()

        series = self.query_ascii_values('FETC?', container=pd.Series)
        if len(series) == 1:
            return series.iloc[0]
        else:
            ii = np.arange(len(series))
            series.index = pd.Index(self.sweep_aperture * ii, name='Time elapsed (s)')
            series.name = 'Power (dBm)'
            return series

    def zero(self):
        with self.overlap_and_block(30):
            self.write('CAL:ZERO:AUTO ONCE')

    def calibrate(self):
        with self.overlap_and_block(10):
            self.write('CAL:AUTO ONCE')

    @contextlib.contextmanager
    def overlap_and_block(self, timeout=None, quiet=False):
        """context manager that sends '*OPC' on entry, and performs
        a blocking '*OPC?' query on exit.

        These SCPI commands instruct the sensor to execute the commands
        concurrently. On exit out of the python context block, execution
        blocks until all of the commands have completed.

        Example::

            with inst.overlap_and_block():
                inst.write('long running command 1')
                inst.write('long running command 2')

        Arguments:
            timeout: maximum time to wait for '*OPC?' reply, or None to use `self.backend.timeout`
            quiet: Suppress timeout exceptions if this evaluates as True

        Raises:
            TimeoutError: on '*OPC?' query timeout
        """

        self._opc = True
        yield
        self._opc = False

        self._await_completion()
        
    def _await_completion(self, timeout: float=None):
        if timeout is None:
            timeout = self.timeout

        # monitoring *ESR? is recommended by the programming manual
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < timeout:
            try:
                register = int(self.backend.query('*ESR?'))
                time.sleep(0.1)
            except pyvisa.errors.VisaIOError as ex:
                raise ex
            else:
                if register & 1:
                    # first bit == operation complete
                    break
        else:
            raise TimeoutError('command failed')

    def _clear(self):
        self.write('*CLS')

    def _event_status_enable(self):
        self.write('*ESE 1')

    def _check_errors(self):
        code, text = self.query('SYST:ERR?').split(',', 1)
        code = int(code)

        if code == 0:
            return
        else:
            raise IOError(code, text[1:-1])


class KeysightU2044XA(KeysightU2000XSeries):
    model = attr.value.str('U2044XA', inherit=True)


class RohdeSchwarzNRPSeries(lb.VISADevice):
    """Coaxial power sensors connected by USB.

    These require the installation of proprietary drivers from the vendor website.
    Resource strings for connections take the form 'RSNRP::0x00e2::103892::INSTR'.
    """

    _FUNCTIONS = ('POW:AVG', 'POW:BURS:AVG', 'POW:TSL:AVG', 'XTIM:POW', 'XTIM:POWer')
    _TRIGGER_SOURCES = ('HOLD', 'IMM', 'INT', 'EXT', 'EXT1', 'EXT2', 'BUS', 'INT1')

    # Instrument state traits (pass command arguments and/or implement setter/getter)
    frequency = attr.property.float(
        key='SENS:FREQ', min=10e6, step=1e-3, label='Hz', help='calibration frequency'
    )
    initiate_continuous = attr.property.bool(
        key='INIT:CONT',
        help='whether to enable triggering, enabling measurement acquisition',
    )
    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    function = attr.property.str(key='SENS:FUNC', case=False, only=_FUNCTIONS)

    @function.setter
    def _(self, value):
        # special case SCPI format: requires quotes around the argument
        self.write(f'SENSe:FUNCtion "{value}"')

    @attr.property.str(key='TRIG:SOUR', case=False, only=_TRIGGER_SOURCES)
    def trigger_source(self):
        """'HOLD: No trigger; IMM: Software; INT: Internal level trigger; EXT2: External trigger, 10 kOhm"""

        # special case - the instrument returns '2' instead of 'EXT2'
        remap = {'2': 'EXT2'}
        source = self.query('TRIG:SOUR?')
        return remap.get(source, default=source)

    trigger_delay = attr.property.float(key='TRIG:DELAY', min=-5, max=10, label='s')
    trigger_count = attr.property.int(key='TRIG:COUN', min=1, max=8192)
    trigger_holdoff = attr.property.float(
        key='TRIG:HOLD',
        min=0,
        max=10,
        label='s',
        help='wait time after trigger before acquisition',
    )
    trigger_level = attr.property.float(
        key='TRIG:LEV',
        min=1e-7,
        max=200e-3,
        label='dBm?',
        help='threshold for power triggering',
    )

    trace_points = attr.property.int(
        key='SENSe:TRACe:POINTs',
        min=1,
        max=8192,
        gets=False,
        help='number of contiguous power samples to acquire per trigger',
    )
    trace_realtime = attr.property.bool(key='TRAC:REAL')
    trace_time = attr.property.float(key='TRAC:TIME', min=10e-6, max=3, label='s')
    trace_offset_time = attr.property.float(
        key='TRAC:OFFS:TIME', min=-0.5, max=100, label='s'
    )
    trace_average_count = attr.property.int(
        key='TRAC:AVER:COUN', min=1, max=65536, label='samples'
    )
    trace_average_mode = attr.property.str(
        key='TRAC:AVER:TCON', only=('MOV', 'REP'), case=False
    )
    trace_average_enable = attr.property.bool(key='TRAC:AVER')

    average_count = attr.property.int(key='AVER:COUN', min=1, max=65536)
    average_auto = attr.property.bool(key='AVER:COUN:AUTO')
    average_enable = attr.property.bool(key='AVER')
    smoothing_enable = attr.property.bool(key='SMO:STAT', gets=False)

    # Local settings traits (leave command unset, and do not implement setter/getter)
    read_termination: str = attr.value.str('\n', inherit=True)

    def preset(self):
        self.write('*PRE')

    def trigger_single(self):
        self.write('INIT')

    def reset(self):
        self.write('*RST')

    def fetch(self) -> typing.Union[SeriesType, float]:
        """Return a single number or pandas Series containing the power readings"""
        series = self.query_ascii_values('FETC?', container=pd.Series)
        if len(series) == 1:
            return float(series.iloc[0])
        series.index = np.arange(len(series)) * (
            self.trace_time / float(self.trace_points)
        )
        series.name = 'Power (dBm)'
        return series

    def fetch_buffer(self):
        """Return a single number or pandas Series containing the power readings"""
        response = self.query('FETC:ARR?').split(',')
        if len(response) == 1:
            return float(response[0])
        else:
            index = None  # np.arange(len(response))*(self.trace_time/float(self.trace_points))
            return pd.to_numeric(pd.Series(response, index=index))

    def setup_trace(
        self,
        frequency: float,
        trace_points: int,
        sample_period: float,
        trigger_level: float,
        trigger_delay: float,
        trigger_source: str,
    ) -> None:
        """establish trace operation.

        Arguments:
            frequency: in Hz
            trace_points: number of points in the trace (perhaps as high as 5000)
            sample_period: in s
            trigger_level: in dBm
            trigger_delay: in s
            trigger_source: 'HOLD: No trigger; IMM: Software; INT: Internal level trigger; EXT2: External trigger, 10 kOhm'
        """

        warnings.warn(
            'setup_trace is deprecated; use RohdeSchwarzNRPTrace instead',
            DeprecationWarning,
        )

        self.reset()
        self.frequency = frequency
        self.function = 'XTIM:POW'
        self.trace_points = trace_points
        self.trace_time = trace_points * sample_period
        self.trigger_level = 10 ** (trigger_level / 10.0)
        self.trigger_delay = trigger_delay  # self.Ts / 2
        self.trace_realtime = True
        self.trigger_source = trigger_source  # 'EXT2'  # Signal analyzer trigger output (10kOhm impedance)
        self.initiate_continuous = False
        self.wait()


class RohdeSchwarzNRP8s(RohdeSchwarzNRPSeries):
    frequency = attr.property.float(inherit=True, max=8e9)


class RohdeSchwarzNRP18s(RohdeSchwarzNRPSeries):
    frequency = attr.property.float(inherit=True, max=18e9)


class PowerTrace_RohdeSchwarzNRP(lb.Rack):
    sensor: RohdeSchwarzNRPSeries

    def setup_trace(
        self,
        frequency: float,
        trace_points: int,
        sample_period: float,
        trigger_level: float,
        trigger_delay: float,
        trigger_source: str,
    ) -> None:
        """establish trace operation.

        Arguments:
            frequency: in Hz
            trace_points: number of points in the trace (perhaps as high as 5000)
            sample_period: in s
            trigger_level: in dBm
            trigger_delay: in s
            trigger_source: 'HOLD: No trigger; IMM: Software; INT: Internal level trigger; EXT2: External trigger, 10 kOhm'
        """

        self.sensor.reset()
        self.sensor.frequency = frequency
        self.sensor.function = 'XTIM:POW'
        self.sensor.trace_points = trace_points
        self.sensor.trace_time = trace_points * sample_period
        self.sensor.trigger_level = 10 ** (trigger_level / 10.0)
        self.sensor.trigger_delay = trigger_delay
        self.sensor.trace_realtime = True
        self.sensor.trigger_source = trigger_source  # 'EXT2'  # Signal analyzer trigger output (10kOhm impedance)
        self.sensor.initiate_continuous = False
        self.sensor.wait()


if __name__ == '__main__':
    from matplotlib import pyplot as plt
    import seaborn as sns

    sns.set(style='ticks')

    # Enable labbench debug messages
    lb.show_messages('info')

    with KeysightU2000XSeries() as sensor:
        print('Connected to ', sensor.identity)

        # Configure
        sensor.preset()
        sensor.frequency = 1e9
        sensor.measurement_rate = 'FAST'
        sensor.trigger_count = 200
        sensor.sweep_aperture = 20e-6
        sensor.trigger_source = 'IMM'
        sensor.initiate_continuous = True

        power = sensor.fetch()

    power.hist(figsize=(6, 2))
    plt.xlabel('Power level')
    plt.ylabel('Count')
    plt.title('Histogram of power sensor readings')



from typing import Union


def path(*subdirs, platform: Union[bool, str] = False) -> str:
    """return a path to a distribution file inside ssmdevices.lib.

    Args:
        platform:
            if False, assume the file is platform-independent;
            if True, use the current platform;
            otherwise, the name of another architecture with format f'darwin-arm64'
    """

    from pathlib import Path
    from platform import system, machine

    if platform is True:
        platform_name = f'{system().lower()}-{machine().lower()}'
        subdirs = (platform_name,) + tuple(subdirs)
    elif isinstance(platform, str):
        subdirs = (platform,) + tuple(subdirs)

    path = Path(__path__[0], *subdirs)

    if platform is not False:
        children = [p for p in path.iterdir() if p.stem.lower() == subdirs[-1].lower()]
        if len(children) == 1:
            path = path/children[0]
        elif len(children) == 0:
            raise IOError(f'no files matched {platform_name} in platform directory {path}')
        elif len(children) > 1:
            raise IOError(f'multiple files matched {platform_name} in platform directory {path}')

    return str(path.relative_to(Path('.').absolute()))


"""
Basic control over the QXDM software.

Author: Paul Blanchard (paul.blanchard@nist.gov)
Edits by Dan Kuester (dkuester@nist.gov)
"""

__all__ = ['QXDM']

import datetime
import os
import time
from xml.etree import ElementTree as ET

import labbench as lb
from labbench import paramattr as attr
import psutil


class QPST(lb.Win32ComDevice):
    PORT_LIST_CODES = dict(
        ue_mode={
            0: 'No phone detected',
            2: 'Download',
            3: 'Diagnostic',
            4: 'Offline and diagnostic',
            5: 'Streaming download',
        }
    )

    API_ATTR_MAP = dict(
        ue_model_number='ModelNumber',
        ue_mode='PhoneMode',
        ue_imei='IMEI',
        ue_esn='ESN',
        ue_build_id='BuildId',
    )

    com_object = attr.value.str('QPSTAtmnServer.Application', inherit=True)

    def open(self):
        # have to do this for every method we want to call
        self.backend._FlagAsMethod('AddPort')
        self.backend._FlagAsMethod('RemovePort')
        self.backend._FlagAsMethod('GetPort')
        self.backend.HideWindow()

    def close(self):
        self.backend.Quit()

    def add_port(self, port):
        """enables the specified port in QPST"""

        self.backend.AddPort(f'COM{port}', 'Python generated port')

        for elapsed in lb.timeout_iter(5):
            try:
                port_list = self.backend.GetPortList
                for i in range(port_list.PhoneCount):
                    if port_list.PortName(i).upper() == f'COM{port}':
                        if port_list.PhoneStatus(i) != 5:
                            raise ValueError(f'qpst detected no phone on COM{port}')
                        ret = {}
                        for key, api_name in self.API_ATTR_MAP.items():
                            func = getattr(port_list, api_name)
                            port_code = func(i)

                            # Remap codes into strings
                            label = self.PORT_LIST_CODES.get(key, {}).get(
                                port_code, port_code
                            )
                            ret[key] = label
                        break

                else:
                    raise TimeoutError(f'qpst timeout adding port COM{port}')
            except ValueError:
                lb.sleep(0.1)
                continue
            else:
                break
        else:
            raise IOError(f'qpst connection failure on COM{port}')

        return ret

    def remove_port(self, port):
        """removes the port from QPST for consistency"""
        self.backend.RemovePort(f'COM{port}')

        for elapsed in lb.timeout_iter(5):
            try:
                port_list = self.backend.GetPortList
                for i in range(port_list.PhoneCount):
                    if port_list.PortName(i).upper() == f'COM{port}':
                        break
                else:
                    break
            except ValueError:
                lb.sleep(0.1)
                continue
            else:
                break
        else:
            raise TimeoutError(f'QXDM disconnect timeout on COM{port}')


class QXDM(lb.Win32ComDevice):
    """QXDM software wrapper"""

    com_object = attr.value.str('QXDM.QXDMAutoApplication', inherit=True)

    resource: int = attr.value.int(
        0, min=0, help='serial port number for the handset connection'
    )
    cache_path: str = attr.value.str(
        default='temp', help='directory for auto-saved isf files'
    )
    connection_timeout: float = attr.value.float(
        default=2, min=0.5, help='connection timeout (s)'
    )

    def open(self):
        #
        self._killall()

        self._connection_info = {}
        self._qpst = QPST()
        self._qpst.open()
        self._connection_info = self._qpst.add_port(self.resource)

        self._start_time = None
        self._window = self.backend.GetAutomationWindow()
        self.cache_path = os.path.abspath(self.cache_path)
        os.makedirs(self.cache_path, exist_ok=True)

        while not self._qpst.isopen:
            lb.sleep(0.1)

        # Disable to prevent undesired data streaming on startup
        try:
            self._set_com_port(None)
        except TimeoutError:
            raise Exception(
                'could not disable UE; does QXDM work if you start it manually?'
            )

    # State traits implemented by command key
    ue_model_number = attr.property.str(help='model number code', key='ue_model_number')
    ue_mode = attr.property.str(help='current state of the phone', key='ue_mode')
    ue_imei = attr.property.str(help='Phone IMEI', key='ue_imei')
    ue_esn = attr.property.str(help='Phone ESN', key='ue_esn')
    ue_build_id = attr.property.str(
        help='Build ID of software on the phone', key='ue_build_id'
    )

    def get_key(self, key, trait_name=None):
        try:
            return self._connection_info[key]
        except KeyError:
            raise lb.DeviceStateError(f"no state information for key '{key}'")

    def close(self):
        def none_func():
            pass

        try:
            f1 = self._qpst.disconnect
        except AttributeError:
            f1 = none_func
            self._logger.debug('QPST already quit')
        try:
            f2 = self._window.QuitApplication
        except AttributeError:
            f2 = none_func
            self._logger.debug('QXDM already quit')

        lb.concurrently(f1, f2)

    def configure(self, config_path, min_acquisition_time=None):
        """Load the QXDM .dmc configuration file at the specified path,
        with adjustments that disable special file output modes like
        autosave, quicksave, and automatic segmenting based on time and
        file size.
        """
        if not os.path.isfile(config_path):
            raise Exception('config_path {} does not exist.'.format(repr(config_path)))
        self._min_acquisition_time = min_acquisition_time

        basename = os.path.splitext(os.path.basename(config_path))[0] + '-live.dmc'
        path_out = os.path.join(self.cache_path, basename)

        tree = ET.parse(config_path)
        root = tree.getroot()
        settings = root.find('Persistence').find('MainFrame').find('ISFSettings')

        # Disable automatic file saving and partitioning
        settings.find('AutoISFSave').text = '0'
        settings.find('QuickISFSave').text = '0'
        settings.find('QueryISFSave').text = '0'
        # These seem to be irrelevant now
        #        settings.find('BaseISFName').text = self.save_base_name
        #        settings.find('ISFFolder').text = self.cache_path
        settings.find('MaxISFSize').text = '0'  # str(self.save_size_limit_MB)
        settings.find('MaxISFDuration').text = '0'
        settings.find('MaxISFDurationFraction').text = '0'  # str(self.save_time_limit)
        settings.find('Advanced').text = '0'

        # Don't care about this any more?
        #        isv_config = root.find('Persistence').find('LoggingView').find('ISVConfig')
        #        isv_config.find('LogFilePath').text = self.cache_path

        tree.write(path_out)
        self._load_config(path_out)
        self._logger.debug('loaded modified configuration at {}'.format(repr(path_out)))

    def save(self, path=None, saveNm=None):
        """Stop the run and save the data in a file at the specified path.
        If path is None, autogenerate with self.cache_path and
        self.data_filename.

        This method is threadsafe.

        :returns: The absolute path to the data file
        """
        if self._start_time is None:
            raise Exception('call start() to acquire data before calling save()')
        if self._min_acquisition_time is not None:
            t_elapsed = time.time() - self._start_time
            if t_elapsed < self._min_acquisition_time:
                lb.sleep(self._min_acquisition_time - t_elapsed)
        # Munge path
        if path is None:
            now = datetime.datetime.now()
            fmt = lb.Host.time_format.replace(' ', '_').replace(':', '')
            timestamp = '{}.{}'.format(now.strftime(fmt), now.microsecond)
            if saveNm is not None:
                path = os.path.join(
                    self.cache_path, '{}-{}.isf'.format(saveNm, timestamp)
                )
            else:
                path = os.path.join(self.cache_path, 'qxdm-{}.isf'.format(timestamp))
        else:
            path = os.path.abspath(path)

        # Stop acquisition
        with lb.stopwatch('saving data'):
            self._set_com_port(None)
            self._wait_for_stop()

            # Save the file
            self._window.SaveItemStore(path)
            self._logger.debug(f'saved to {path}')

        self._start_time = None
        return path

    def start(self, wait=True):
        """Start acquisition, optionally waiting to return until
        new data enters the QXDM item store.
        """
        t0 = time.time()
        self._clear()
        self._set_com_port(self.resource)
        if wait:
            self._wait_for_start()
        self._logger.debug('running after setup for {}s'.format(time.time() - t0))

        self._start_time = time.time()

    # Bare wrapper methods for the low level QXDM COM API
    def _get_com_port(self):
        return self._window.getCOMPort()

    def _load_config(self, path):
        self._window.LoadConfig(path)

    @attr.property.str(sets=False, cache=True)
    def version(self):
        """QXDM application version"""
        _window = self.backend.GetAutomationWindow()
        if not self.isopen:
            raise lb.DeviceNotReady('need to connect to get application version')
        version = _window.AppVersion
        return version

    def _get_server_state(self):
        state = self._window.GetServerState()
        if state == 0xFFFFFFFF:
            raise Exception('server state is in error')
        return state

    def _get_item_count(self):
        return self._window.GetItemCount()

    def _killall(self):
        """kills all instances of qpst, qxdm, and atmnserver"""
        for pid in psutil.pids():
            try:
                proc = psutil.Process(pid)
                for target in 'qpst', 'qxdm', 'atmnserver':
                    if proc.name().lower().startswith(target.lower()):
                        self._logger.debug(
                            'killing zombie process {}'.format(proc.name())
                        )
                        proc.kill()
            except psutil.NoSuchProcess:
                pass

    def _set_com_port(self, com_port: int):
        """connects to the handset at the specified serial port.

        Blocks until QXDM confirms the phone is connected.

        Arguments:
            com_port (int): serial port number, COMn; 0 or None to disable

        Raises:
            TimeoutError: after waiting `self.connection_timeout` for connection
            IOError: when the QXDM COM API returns code -1

        """
        if com_port is None:
            com_port = 0
        if int(com_port) > 0:
            self._connection_info = self._qpst.add_port(self.resource)

        try:
            code = None

            for elapsed in lb.timeout_iter(self.connection_timeout):
                # try it
                ret = int(self._window.setCOMPort(com_port))

                # validate
                if ret == -1:
                    raise IOError('Connection error')
                actual = self._get_com_port()
                if str(actual) == str(com_port):
                    break
                else:
                    code = self._window.GetServerState()
                    if code == 0xFFFFFFFF:
                        raise Exception('Connection error')
                    lb.sleep(0.1)
            else:
                if com_port:
                    raise TimeoutError(f'handset connect to port {actual} timed out')
                else:
                    raise TimeoutError(f'handset disconnect to port {actual} timed out')

            if int(com_port) > 0:
                self._logger.debug(f'connect to port {actual} took {elapsed:0.2f}s')
            else:
                self._logger.debug(
                    f'disconnect from port {actual} took {elapsed:0.2f}s'
                )

        finally:
            if int(com_port) == 0:
                self._qpst.remove_port(self.resource)

    def reconnect(self):
        self._logger.info('reinitializing')
        self.close()
        self.open()

    def _clear(self, timeout=20):
        """clears the buffer of data.

        TODO: Depending on if QXDM is already running, wait for the item
        count store to start increasing again?
        """
        start = self._get_item_count()
        for item in 'Item view', 'Filtered view':
            if not self._window.ClearViewItems(item):
                if item == 'Item view':
                    self._logger.error('failed to clear view ' + repr(item))

        # Block until the item store is clear or timeout
        for elapsed in lb.timeout_iter(timeout):
            count = self._get_item_count()
            if count < start or count < 10:
                break
            lb.sleep(0.05)
        else:
            remaining = self._get_item_count()
            raise TimeoutError(
                f'QXDM item store clear timed out with {remaining} items'
            )

        self._logger.debug(f'cleared item store buffer in {elapsed:0.2f}s')

    def _wait_for_stop(self, timeout=10):
        """blocks until the QXDM item store stops growing"""
        prev = self._get_item_count()
        for elapsed in lb.timeout_iter(timeout):
            lb.sleep(0.25)
            new = self._get_item_count()
            if new == prev:
                break
            else:
                prev = new
        else:
            raise TimeoutError('timeout waiting for qxdm to buffer data')

    def _wait_for_start(self, timeout=10, min_rows=10):
        """blocks until the item store has grown by at least `min_rows`"""

        start = self._get_item_count()
        for elapsed in lb.timeout_iter(timeout):
            lb.sleep(0.05)
            if self._get_item_count() != start:
                break
        else:
            raise Exception('timeout waiting for qxdm to start acquisition')
        #        lb.sleep(1)
        self._logger.debug(f'item store grew after {elapsed:0.2f}s')

if __name__ == '__main__':
   import labbench as lb
   import inspect

   lb.show_messages('debug')

   # Connect to application
   with QXDM(8, cache_path=r'C:\Python Code\potato', concurrency=False) as qxdm:
       mod = inspect.getmodule(qxdm.backend._FlagAsMethod).__name__
       print(repr(qxdm.backend),dir(qxdm.backend))
       qxdm.configure(r'180201_QXDMConfig.dmc')
       for i in range(1):
           qxdm.start()
           lb.sleep(10)
           qxdm.save(r'junk-{}.isf'.format(i))


import psutil
import socket
import re

INTERFACE_FIELDS = ('interface', 'ip_address', 'physical_address', 'ipv6_address')


def list_network_interfaces(by='interface'):
    f"""Try to look up the IP address corresponding to the network interface
    referred to by the OS with the name `iface`.

    If the interface does not exist, the medium is disconnected, or there
    is no IP address associated with the interface, raise `ConnectionError`.

    :param by: the field to use as the key in the return dictionary. one of {INTERFACE_FIELDS}

    """
    if by not in INTERFACE_FIELDS:
        raise ValueError(f"by '{by}' is not one of {INTERFACE_FIELDS}")

    ret = {}
    for name, if_structs in psutil.net_if_addrs().items():
        iface = dict(interface=name)
        for family_info in if_structs:
            if family_info.family is socket.AF_INET:
                iface['ip_address'] = family_info.address
                iface['ip_netmask'] = family_info.netmask
                iface['ip_broadcast'] = family_info.broadcast
            elif family_info.family is socket.AF_INET6:
                iface['ipv6_address'] = family_info.address
                iface['ipv6_netmask'] = family_info.netmask
                iface['ipv6_broadcast'] = family_info.broadcast
            elif family_info.family is psutil.AF_LINK:
                iface['physical_address'] = family_info.address.replace(
                    '-', ':'
                ).lower()

        if by in iface:
            ret[iface[by]] = iface

    return ret


def network_interface_info(resource):
    """Try to look up the IP address of a network interface by its name
    or MAC (physical) address.

    If the interface does not exist, the medium is disconnected, or there
    is no IP address associated with the interface, raise `ConnectionError`.
    """
    if re.match(
        r'([0-9a-f]{2}:){5}[0-9a-f]{2}',
        resource.lower().replace('-', ':'),
        re.IGNORECASE,
    ):
        # it's a physical address
        addrs = list_network_interfaces('physical_address')
        resource = resource.lower().replace('-', ':')
    else:
        addrs = list_network_interfaces('interface')

    # Check whether the interface exists
    if resource not in addrs:
        available = ', '.join(addrs.keys())
        msg = (
            f'requested interface for "{resource}" but only ({available}) are available'
        )
        raise ConnectionError(msg)

    return addrs[resource]


def get_ipv4_occupied_ports(ip):
    return {
        conn.laddr[1]
        for conn in psutil.net_connections(kind='inet4')
        if ip in conn.laddr
    }


def get_ipv4_address(resource):
    """Try to look up the IP address of a network interface by its name
    or MAC (physical) address.

    If the interface does not exist, the medium is disconnected, or there
    is no IP address associated with the interface, raise `ConnectionError`.
    """
    info = network_interface_info(resource)

    # Check whether it's up, which is necessary for a good IP address
    if not psutil.net_if_stats()[info['interface']].isup:
        raise ConnectionError(
            f"the network interface {info['interface']} ({info['physical_address']}) is disabled or disconnected"
        )

    # lookup and return the address
    if 'ip_address' not in info:
        raise ConnectionError(
            f"no ipv4 address associated with interface '{info['interface']}' ({info['physical_address']})"
        )

    return info['ip_address']


def find_free_port():
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


if __name__ == '__main__':
    print(list_network_interfaces('physical_address'))



"""
Drivers for USB peripherals

:author: Dan Kuester <daniel.kuester@nist.gov>, Andre Rosete <andre.rosete@nist.gov>
"""

import labbench as lb
from labbench import paramattr as attr

__all__ = ['SwiftNavPiksi']


class SwiftNavPiksi(lb.SerialLoggingDevice):
    baud_rate: int = attr.value.int(
        default=1000000,
        min=1,
    )


if __name__ == '__main__':
    import labbench as lb
    from labbench import paramattr as attr

    lb.debug_to_screen(lb.DEBUG)
    with SwiftNavPiksi.from_hwid(r'USB VID:PID=0403:6014 SER=5') as piksi:
        piksi.start()
        lb.sleep(5)
        piksi.stop()
        result = piksi.fetch()
        print('Received:\n', result)



# Including this file ensures this directory is installed by setuptools


def default_path(cls):
    """Return a module imported from the provided path to a dll
    that contains a .NET library called `libname.`

    :param cls: the class

    :param module_: the module where the config lives
    """

    global __path__
    import os

    mod_split = cls.__module__.split('.', 1)[-1]
    rel_path = (mod_split + '.' + cls.__name__).replace('.', os.path.sep)
    return os.path.join(__path__[0], rel_path)



from .attenuators import *
from .base_station_emulators import *
from .gps_simulator import *
from .networking import *
from .network_analyzers import *
from .oscilloscopes import *
from .power_sensors import *
from .power_supplies import *
from .signal_analyzers import *
from .signal_generators import *
from .switches import *
from .motors import *
